/*
 * frc        start the bootup/shutdown scripts in rc?.d
 *       
 * features   parallel startup, colourisation, interaction
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sysexits.h>
#include <termios.h>
#include <signal.h>
#include <pwd.h>
#include <shadow.h>
#include <crypt.h>

#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef USE_IDSA
#include <idsa.h>
#else
#include <syslog.h>
#endif

#ifdef USE_CURSES
#include <ncurses.h>
#include <term.h>
#endif

/****************************************************************************/

#define FRC_BUFFER     512
#define FRC_NAME       "frc"
#define FRC_BASE       "/etc"
#define FRC_SHELL      "/bin/sh"
#define FRC_SPACE      "  "
#define FRC_TIMEOUT     30

#define SHOW_STDERR    0x01   /* display stderr of child */
#define SHOW_STDOUT    0x02   /* display stdout of child */
#define SHOW_TRANSIENT 0x04   /* show "busy" messages */
#define SHOW_STATUS    0x08   /* show status */
#define SHOW_TOTAL     0x10   /* show last summary */
#define SHOW_BANNER    0x20   /* show splash/help */
#define SHOW_COLOURS   0x40   /* in technicolour */
#define SHOW_VERBOSE   0x80   /* show misc gunk */

#define DECO_NONE      0x00
#define DECO_RED       0x01
#define DECO_GREEN     0x02
#define DECO_YELLOW    0x03
#define DECO_COLOURS   0x0f
#define DECO_ERASE     0x10

#ifndef FRC_VERSION
#define FRC_VERSION "unknown"
#endif

#define CMD_ALT        0x01
#define CMD_YES        0x02
#define CMD_NO         0x04

#define STATE_OPEN     0x01   /* user can request interaction */
#define STATE_AUTH     0x02   /* authentication ok */
#define STATE_ASK      0x10   /* should ask if start */
#define STATE_KILL     0x20   /* should ask for kill */

#define REQ_START    "start"
#define REQ_STOP     "stop"
#define REQ_KILL     "kill"

/****************************************************************************/

struct task{
  pid_t t_pid;
  int t_err;
  int t_out;
  int t_mode;
  char *t_name;
};

struct decoration{
  int d_erase;
  int d_margin;
  int d_show;
  char *d_colour;
  char *d_reset;
  char *d_bold;
  char *d_clear;
};

struct decoration frc_deco;

struct termios    frc_save;  /* sane settings at startup */
struct termios    frc_want;  /* the ones we normally use */

char             *frc_base;  /* base directory for scripts */
int              frc_login;  /* make sushell a login shell ? */
int             frc_setsid;  /* give each child script its own session ? */
volatile int     frc_state;  /* authentication state */
int            frc_seconds;  /* how many second credit ? */

#define should_show(p) (frc_deco.d_show & p)

/****************************************************************************/

int  run_list(char level, char **runv, int runc, struct task *taskv, int taskm, char *req);
int  run_concurrent(char level, struct task *taskv, int taskc, char *req);
int  activate_child(char level, struct task *task, char *req);

DIR *runlevel_open(char level);

int  zadelete(char **vec, int cnt, char *match);

int  vec_insert(char ***ptr, int cnt, char *str);
int  vec_sort(char **vec, int cnt);
int  vec_compare(const void *alpha, const void *beta);
void vec_free(char **vec, int cnt);
void vec_dump(char *str, char **vec, int cnt, FILE *fp);

int  test_interactive(char level, char *req, char *service, int state);
void run_shell();

void trms_quit();
void trms_reinit();
void trms_init();

void fatal_error(int code, char *message, ...);

void display_list(char *req, struct task *taskv, int taskc);
void display_child(int mode, char *prefix, int fd, int display);
void display_ok(char *message, ...);
void display_error(char *message, ...);
void display_transient(char *message, ...);
void display_plain(char *message, ...);
void display(int mode, char *prefix, char *message, ...);

void deco_vdisplay(int mode, char *prefix, char *message, va_list args);
void deco_display(int mode, char *prefix, char *message);
void deco_change(int show);
void deco_compute(char **vec, int cnt);

char *zaprefix(char *str);
int   chprefix(char *a, char *b);

/****************************************************************************/

char *zaprefix(char *str)
{
  if((str[0] == 'K') || (str[0] == 'S')){
    str++;
  }

  while(isdigit(str[0])){
    str++;
  }

  return str;
}

int chprefix(char *a, char *b)
{
  int i;

  i=1;
  while(isdigit(a[i])||isdigit(b[i])){
    if(a[i]!=b[i]){
      return 1;
    }
    i++;
  }
  return 0;
}

/****************************************************************************/

char *ingify(char *s){
  if(strcmp("stop", s) == 0){
    return "ping" ;
  } else {
    return "ing";
  }
}

char *edify(char *s){
  if(strcmp("stop", s) == 0){
    return "ped" ;
  } else {
    return "ed";
  }
}

/****************************************************************************/

static void handle_interrupt(int s)
{
  frc_state |= (STATE_ASK | STATE_KILL); /* enable all interaction */
}

/****************************************************************************/

static void deco_setup(int show)
{
  int t;

  frc_deco.d_erase  = 0;
  frc_deco.d_margin = strlen(FRC_NAME)+1;

  frc_deco.d_show   = show;
  frc_deco.d_colour = NULL;
  frc_deco.d_reset  = NULL;
  frc_deco.d_bold   = NULL;
  frc_deco.d_clear  = NULL;

#ifdef USE_CURSES
  if (setupterm(NULL, STDOUT_FILENO, &t) != OK) {
    return;
  }

  frc_deco.d_clear = tigetstr("el");
  if(frc_deco.d_clear == NULL){
    frc_deco.d_show &= (~SHOW_TRANSIENT);
  }

  frc_deco.d_reset = tigetstr("sgr0");
  if(frc_deco.d_reset){
    frc_deco.d_colour = tigetstr("setaf");
    frc_deco.d_bold = tigetstr("bold");
    if(frc_deco.d_bold == NULL){
      frc_deco.d_reset = NULL;
    }
  }

#ifdef DEBUG
  fprintf(stderr, "colour string=%s\n", frc_deco.d_colour);
#endif
#endif
}

void deco_compute(char **vec, int cnt)
{
  int i,n;

  frc_deco.d_margin = 0;

  for(i=0;i<cnt;i++){
    n = strlen(zaprefix(vec[i]));
    if(n > frc_deco.d_margin){
      frc_deco.d_margin = n;
    }
  }
  frc_deco.d_margin++;
}

void deco_change(int show)
{
  frc_deco.d_show = show;
}

void deco_display(int mode, char *prefix, char *message)
{
  int i, m, s, x;

  if(message){


    s = 1;
    i = 0;
    m = strlen(message);

    while(i < m){
      if(s){
        s = 0;

        if(frc_deco.d_erase){ /* clear line if needed */
          fputc('\r', stdout);
          putp(frc_deco.d_clear);
        }

        if(prefix){ /* print prefix, in technicolour */
          if(frc_deco.d_bold){
            putp(frc_deco.d_bold);
            if(should_show(SHOW_COLOURS)){
              if(mode & DECO_COLOURS){
                putp(tparm(frc_deco.d_colour, mode & DECO_COLOURS));
              }
            }
          }
          x = strlen(prefix);
          while(x < frc_deco.d_margin){
            fputc(' ', stdout);
            x++;
          }
          fputs(prefix, stdout);
          if(frc_deco.d_reset){
            putp(frc_deco.d_reset);
          }
          fputc(':', stdout);
          fputc(' ', stdout);
        } 
      } else {
        if(message[i] == '\n'){
          s = 1;
        } else {
          fputc(message[i], stdout);
        }

        if(s || ((i+1) == m)){ /* if newline or end of string shedule erase */
          if(mode & DECO_ERASE){
            frc_deco.d_erase = 1;
          } else {
            frc_deco.d_erase = 0;
            fputc('\n', stdout);
          }
        }
        i++;
      }
    }

    fflush(stdout);
  }
}

void deco_vdisplay(int mode, char *prefix, char *message, va_list args)
{
  char buffer[FRC_BUFFER];

  vsnprintf(buffer, FRC_BUFFER - 1, message, args);
  buffer[FRC_BUFFER - 1] = '\0';

  deco_display(mode, prefix, buffer);
}

/****************************************************************************/

void display(int mode, char *prefix, char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(mode, prefix, message, args);
  va_end(args);
}

void display_plain(char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(DECO_NONE, FRC_NAME, message, args);
  va_end(args);
}

void display_transient(char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(DECO_YELLOW | DECO_ERASE, FRC_NAME, message, args);
  va_end(args);
}

void display_error(char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(DECO_RED, FRC_NAME, message, args);
  va_end(args);
}

void display_ok(char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(DECO_GREEN, FRC_NAME, message, args);
  va_end(args);
}

void fatal_error(int code, char *message, ...)
{
  va_list args;

  va_start(args, message);
  deco_vdisplay(DECO_RED, FRC_NAME, message, args);
  va_end(args);

  trms_quit();

  exit(code);
}

void display_child(int mode, char *prefix, int fd, int display)
{
  char buffer[FRC_BUFFER];
  int rr;

  rr = read(fd, buffer, FRC_BUFFER - 1);
  if(rr > 0){
    if(buffer[rr-1] != '\n'){
      buffer[rr++]='\n';
    }
    buffer[rr] = '\0';
    if(display){
      deco_display(mode, prefix, buffer);
    }
  }
}

void display_list(char *req, struct task *taskv, int taskc)
{
  int m, i;
  char *str;
 
  for(i=0,m=0;i<taskc;i++){
    if(taskv[i].t_pid != (-1)){
      m+=strlen(zaprefix(taskv[i].t_name))+1;
    }
  }

  str = malloc(m*sizeof(char));
  if(str){
    for(i=0,m=0;i<taskc;i++){
      if(taskv[i].t_pid != (-1)){
        strcpy(str+m, zaprefix(taskv[i].t_name));
        m+=strlen(zaprefix(taskv[i].t_name));
        str[m]=' ';
        m++;
      }
    }
    if(m){
      str[m-1]='\0';
    }
    display_transient("%s%s %s ...\n", req, ingify(req), str);
    free(str);
  }
}

/****************************************************************************/

void trms_init()
{
#ifdef DEBUG
  char *term;
  int i;

  for(i=0; i<3 ;i++){
    term = ttyname(i);
    display_plain("fd=%d, terminal=%s\n", i, term ? term : "none");
  }
#endif
  
  if(frc_state & STATE_OPEN){
    /* WARNING: weirdness, make sure we have a controlling tty */
    ioctl(STDIN_FILENO, TIOCSCTTY, 1);
  }

  if(tcgetattr(STDIN_FILENO, &frc_save)){
    display_error("unable to acquire terminal settings: %s\n", strerror(errno));
  } else {

    memcpy(&frc_want, &frc_save, sizeof(struct termios));

#ifdef DEBUG
    display_plain("interrupt=%d stop=%d\n", frc_want.c_cc[VINTR], frc_want.c_cc[VSTOP]);
#endif

    frc_want.c_iflag &= ~(IXON | IXOFF | IXANY);

    frc_want.c_oflag |=  (OPOST | ONLCR);
    frc_want.c_cflag |=  (CREAD | CLOCAL);

    frc_want.c_lflag &= ~(ICANON | ECHO);
    frc_want.c_lflag |=   ISIG;

    frc_want.c_cc[VINTR] = 3;
    frc_want.c_cc[VMIN]  = 1;
    frc_want.c_cc[VTIME] = 0;


    if(tcsetattr(STDIN_FILENO, 0, &frc_want)){
      display_error("unable to change terminal settings: %s\n", strerror(errno));
    }
  }
}

void trms_reinit()
{
  if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &frc_want)){
    display_error("unable to restore terminal settings: %s\n", strerror(errno));
  }
}

void trms_quit()
{
  if(tcsetattr(STDIN_FILENO, TCSAFLUSH, &frc_save)){
    display_error("unable to restore terminal settings: %s\n", strerror(errno));
  }
}

/****************************************************************************/

int no_action(int state)
{
  if(!(state & frc_state)){ /* no request for interaction */
    return 1;
  }

  return 0;
}

#define SEC_BUFFER 32

int no_auth()
{
  char  buffer[FRC_BUFFER];
  char seconds[SEC_BUFFER];
  struct passwd       *pwd;
  struct spwd         *shd;
  char                *enc;
  fd_set               fsr;
  struct timeval        tv;
  int                rr,br;

  if(!(frc_state & STATE_OPEN)){ /* no interaction permitted */
    return 1;
  }

  if(frc_state & STATE_AUTH){ /* yup, allowed */
    return 0; 
  }

  /* FIXME: ask and check for password */
  pwd = getpwnam("root");
  if(pwd && strcmp("x", pwd->pw_passwd)){
    enc = pwd->pw_passwd;
  } else {
    shd = getspnam("root");
    if(shd){
      enc = shd->sp_pwdp;
    } else {
      display_error("could not find root account: %s\n", strerror(errno));
      return 1;
    }
  }

  br = 0;
  tv.tv_sec = 1;
  tv.tv_usec = 0;

  while(frc_seconds){
    FD_ZERO(&fsr);
    FD_SET(STDIN_FILENO, &fsr);

    if(frc_seconds > 0){
      snprintf(seconds, SEC_BUFFER-1, "%d", frc_seconds);
      seconds[SEC_BUFFER-1] = '\0';
    }
    display(DECO_RED | DECO_ERASE, (frc_seconds > 0) ? seconds : NULL, "root password: ");

    if(select(STDIN_FILENO + 1, &fsr, NULL, NULL, (frc_seconds > 0) ? &tv : NULL) > 0){
      rr = read(STDIN_FILENO, buffer + br, 1);
      switch(rr){
        case -1 :
          br = 0;
          display_error("read error: %s\n", strerror(errno));
          break;
        case  0 :
          /* WARNING: logic assumes that timeout only set for nonessential prompts */
          if(frc_seconds > 0){
            return 1;
          }
          br = 0;
          break;
        case 1 :
          switch(buffer[br]){
            case '\r' :
            case '\n' :
              buffer[br] = '\0';
              if(br > 0){
                if(strcmp(crypt(buffer, enc), enc) == 0){
                  frc_state |= STATE_AUTH ;
                  return 0;
                }
              } else {
                if(enc[0] == '\0'){
                  frc_state |= STATE_AUTH ;
                  return 0;
                }
              }
              display_error("incorrect password\n");
              br = 0;
              break;
            default :
              br++;
              if(br >= FRC_BUFFER){
                display_error("disgustingly long password\n");
                br = 0;
              }
              break;
          }
          break;
      }
    }

    if(tv.tv_usec < 420000){
      tv.tv_sec = 1;
      tv.tv_usec = 0;
      if(frc_seconds > 0){
        frc_seconds--;
      }
    }
  }

  return 1;
}

void run_shell()
{
  pid_t pid;
  int result, status;
  struct passwd *pw;
  char *shell, *name, *home;
  char buffer[FRC_BUFFER];

  /* WARNING: disable any timeout for emergencies */
  frc_seconds = (-1);

  if(no_auth()){
    return;
  }

  pw = getpwnam("root");
  if(pw){
    shell = pw->pw_shell;
    home = pw->pw_dir;
  } else {
    display_error("unable to determine details of root: %s\n", strerror(errno));
    shell = FRC_SHELL;
    home = "/";
  }

  if(frc_login){
    snprintf(buffer, FRC_BUFFER-1, "-%s", shell);
    buffer[FRC_BUFFER-1]='\0';
    name = buffer;
  } else {
    name = shell;
  }

  pid = fork();
  switch(pid){
    case -1 : 
      display_error("unable to spawn child: %s\n", strerror(errno));
      result = -1;
      break;
    case 0 :
      /* stuff below here resembles code of Sysvinit by Miquel van Smoorenburg */
      /* in order to make the shell similar to one spawed by sulogin */

      setenv("HOME", home, 1);
      setenv("SHELL", shell, 1);
      setenv("LOGNAME", "root", 1);
      setenv("USER", "root", 1);
      chdir(home);

      fputc('\n', stdout);

      trms_quit();

      signal(SIGINT, SIG_IGN);

      signal(SIGTERM, SIG_DFL);
      signal(SIGTSTP, SIG_DFL);
      signal(SIGQUIT, SIG_DFL);

      execl(shell, name, NULL);
      display_error("unable to execute %s: %s\n", shell, strerror(errno));

      if(strcmp(shell, FRC_SHELL)){
        setenv("SHELL", FRC_SHELL, 1);
        execl(FRC_SHELL, FRC_SHELL, NULL);
        display_error("unable to execute %s: %s\n", FRC_SHELL, strerror(errno));
      }

      exit(1);
      break;
    default :
      if(waitpid(pid, &status, 0) != pid){
        display_error("unable to collect child, continuing anyway\n");
      }
      trms_reinit();
      break;
  }
}

int test_interactive(char level, char *req, char *service, int state)
{
  int cmd;

#ifdef DEBUG
fprintf(stderr, __FUNCTION__ ": state=0x%02x\n", frc_state);
#endif

  if(no_action(state) || no_auth()){
    return CMD_ALT;
  }

  while(1){
    display_transient("%s %s ?\n", req, service);
    cmd = getchar();
    switch(cmd){
      case 'c' :
        frc_state = STATE_OPEN | STATE_AUTH ;
        return CMD_ALT;
        break;
      case 'y' :
        /* WARNING: note that STATE_KILL is gone, but ASK there */
        frc_state = STATE_OPEN | STATE_AUTH | STATE_ASK; 
        return CMD_YES;
        break;
      case 'n' :
        frc_state = STATE_OPEN | STATE_AUTH | STATE_ASK;
        return CMD_NO;
        break;
      case 'h' :
      case '?' :
        display(DECO_YELLOW, "c", "continue noninteractive startup\n");
        display(DECO_YELLOW, "y", "%s service\n", req);
        display(DECO_YELLOW, "n", "do not %s service\n", req);
        display(DECO_YELLOW, "s", "run shell and return to this prompt\n");
        break;
      case 's' :
      case '!' :
        run_shell();
        break;
      case '\n' :
      case '\r' :
        break;
      default :
        display_error("unknown command %c - try ? for help\n", cmd);
        break;
    }
  }

  /* should not happen */
  return CMD_ALT;
}

/****************************************************************************/

#ifdef DEBUG
void vec_dump(char *str, char **vec, int cnt, FILE *fp)
{
  int i;
  
  for(i=0;i<cnt;i++){
    fprintf(fp, "%s: [%d]=<%s>\n", str, i, vec[i]);
  }
}
#endif

int vec_insert(char ***ptr, int cnt, char *str)
{
  char **vec;
  char *nx;

  nx = strdup(str);
  if(nx==NULL){
    display_error("unable to copy string: ignoring service %s\n", str);
    return cnt;
  }

  vec = realloc(*ptr, (cnt+1)*sizeof(char *));
  if(vec==NULL){
    display_error("unable resize %d bytes: ignoring service %s\n", (cnt+1)*sizeof(char *), str);
    free(nx);
    return cnt;
  }

  *ptr=vec;
  vec[cnt]=nx;

#ifdef DEBUG
  fprintf(stderr, "str=%s, cnt=%d, ***=%p, vec=%p, &vec[cnt]=%p, nx=%p\n", str, cnt, ptr, vec, &vec[cnt], nx);
#endif

  cnt++;
  return cnt;
}

int vec_compare(const void *alpha, const void *beta)
{
  return strcmp(*((char **)alpha), *((char **)beta));
}

int vec_sort(char **vec, int cnt)
{
#ifdef DEBUG
  fprintf(stderr, "sort: vec=%p\n", vec);
#endif

  qsort((void *)vec, cnt, sizeof(char *), &vec_compare);

  return 0;
}

void vec_free(char **vec, int cnt)
{
  int i;

  if(vec){
    for(i=0;i<cnt;i++){
      free(vec[i]);
      vec[i]=NULL;
    }
    free(vec);
  }
}

/****************************************************************************/

int zadelete(char **vec, int cnt, char *match)
{
  int i;
  char *zap;

  zap = zaprefix(match);

#ifdef DEBUG
fprintf(stderr, __FUNCTION__ ": removing %s\n", match);
#endif

  for(i=0;(i<cnt)&&strcmp(zaprefix(vec[i]), zap);i++);

  if(i<cnt){
    free(vec[i]);
    i++;
    while(i<cnt){
      vec[i-1]=vec[i];
      i++;
    }
    cnt--;
  }

#ifdef DEBUG
  vec_dump("delete", vec, cnt, stderr);
#endif

  return cnt;
}

/****************************************************************************/

DIR *runlevel_open(char level)
{
  char directory[FRC_BUFFER];
  DIR *dirh;
  int len;

  /* FIXME: check string */
  len=snprintf(directory, FRC_BUFFER, "%s/rc%c.d", frc_base, level);
  directory[FRC_BUFFER-1]='\0';

  if((len > FRC_BUFFER) || (len < 0)){
    fatal_error(EX_DATAERR, "unable to generate directory from base %s\n", frc_base);
  }

#ifdef DEBUG
  fprintf(stderr, "%s: directory=%s\n", __FUNCTION__, directory);
#endif

  dirh = opendir(directory);
  if(dirh == NULL){
    if(errno == ENOENT){
      display_error("%s does not exist\n", directory);
    } else {
      fatal_error(EX_OSERR, "unable to open %s: %s\n", directory, strerror(errno));
    }
  }

  return dirh;
}

/****************************************************************************/

int activate_child(char level, struct task *task, char *req)
{
  char buffer[FRC_BUFFER];
  int len;
  int errp[2], outp[2], ifd;

  if(pipe(errp)){
    display_error("unable to construct error pipe for %s\n", task->t_name);
    return 1;
  }

  if(pipe(outp)){
    display_error("unable to construct output pipe for %s\n", task->t_name);
    close(errp[0]);
    close(errp[1]);
    return 1;
  }

  task->t_pid = fork();
  switch(task->t_pid){
    case -1:
      display_error("unable to fork process for %s: %s\n", task->t_name, strerror(errno));
      close(errp[0]);
      close(errp[1]);
      close(outp[0]);
      close(outp[1]);
      return 1;
      break;
    case 0 :
      close(errp[0]);
      close(outp[0]);

      signal(SIGINT, SIG_IGN);
      signal(SIGTERM, SIG_DFL);

      /* rather ignore signals from tty */
      /* 
      signal(SIGTSTP, SIG_DFL);
      signal(SIGQUIT, SIG_DFL);
      */

      if(errp[1] != STDERR_FILENO){
        if(dup2(errp[1], STDERR_FILENO) != STDERR_FILENO) {
          exit(EX_OSERR);
        }
        close(errp[1]);
      }

      if(outp[1] != STDOUT_FILENO){
        if(dup2(outp[1], STDOUT_FILENO) != STDOUT_FILENO) {
          exit(EX_OSERR);
        }
        close(outp[1]);
      }

      ifd = open("/dev/null", O_RDONLY);
      if(ifd != (-1)){
        if(ifd != STDIN_FILENO){
          if(dup2(ifd, STDIN_FILENO) != STDIN_FILENO) {
            exit(EX_OSERR);
          }
          close(ifd);
        }
      } else {
        close(STDIN_FILENO);
      }

      if(frc_setsid){
        setsid();
      }

      len=snprintf(buffer, FRC_BUFFER, "%s/rc%c.d/%s", frc_base, level, task->t_name);
      buffer[FRC_BUFFER-1]='\0';
      if((len > FRC_BUFFER) || (len < 0)){
        fprintf(stderr, "unable to construct path of %s\n", task->t_name);
        exit(EX_DATAERR);
      }

      execl(buffer, zaprefix(task->t_name), req, NULL);

      fprintf(stderr, "unable to execute %s: %s\n", buffer, strerror(errno));
      exit(EX_OSERR);

      break;
    default :
      task->t_err=errp[0];
      task->t_out=outp[0];
      close(errp[1]);
      close(outp[1]);
      break;
  }
  return 0;
}

int run_concurrent(char level, struct task *taskv, int taskc, char *req)
{
  fd_set fsr;
  pid_t pid;
  struct timeval tv;
  int status, code, result, taskh, mfd, xfd, sr, i;

  result = 0;

  if(should_show(SHOW_TRANSIENT)){ /* show list of children to be started */
    display_list(req, taskv, taskc);
  }
  
  for(i=0;i<taskc;i++){ /* create children */
    result+=activate_child(level, &taskv[i], req);
  }

  for(taskh=taskc;taskh>0;){ /* while children to run */
    FD_ZERO(&fsr);
    mfd = (-1);

    if(should_show(SHOW_TRANSIENT)){ /* show list of children still starting */
      display_list(req, taskv, taskc);
    }

    for(i=0;i<taskc;i++){
      if(taskv[i].t_pid != (-1)){
        FD_SET(taskv[i].t_err, &fsr);
        FD_SET(taskv[i].t_out, &fsr);
        xfd = (taskv[i].t_err > taskv[i].t_out) ? taskv[i].t_err : taskv[i].t_out;
        if(xfd > mfd){
          mfd = xfd;
        }
      }
    }

    mfd++;
    tv.tv_sec = 1;
    tv.tv_usec = 0;
    sr = select(mfd, &fsr, NULL, NULL, &tv);
    for(i=0;i<taskc;i++){
      if(taskv[i].t_pid != (-1)){
        switch(test_interactive(level, REQ_KILL, taskv[i].t_name, STATE_KILL)){
          case CMD_ALT :
            i=taskc;
            break;
          case CMD_YES :
            display_transient("sending %s a TERM signal\n", zaprefix(taskv[i].t_name));
            kill(taskv[i].t_pid, SIGTERM);
            sleep(1);
            display_transient("sending %s the KILL signal\n", zaprefix(taskv[i].t_name));
            sleep(1);
            kill(taskv[i].t_pid, SIGKILL);
            break;
          case CMD_NO :
            break;
        }
      }
    }
    if(sr <= 0){
      continue;
    }

    for(i=0;i<taskc;i++){
      if(taskv[i].t_pid != (-1)){
        if(FD_ISSET(taskv[i].t_err, &fsr)) {
          display_child(DECO_RED, zaprefix(taskv[i].t_name), taskv[i].t_err, should_show(SHOW_STDERR));
        }
        if(FD_ISSET(taskv[i].t_out, &fsr)) {
          display_child(DECO_YELLOW, zaprefix(taskv[i].t_name), taskv[i].t_out, should_show(SHOW_STDOUT));
        }
      }
    }

    while((pid = waitpid(-1, &status, WNOHANG)) > 0) {   /* collect finished children */
      for(i=0;i<taskc;i++){
        if(taskv[i].t_pid == pid){
          if (WIFEXITED(status)) {
            code = WEXITSTATUS(status);
            switch(code){
              case 0 :
                if(should_show(SHOW_STATUS)){
                  display(DECO_GREEN, zaprefix(taskv[i].t_name), "%s[%s ok]\n", FRC_SPACE, req);
                }
                break;
              case 1 :
                if(should_show(SHOW_STATUS)){
                  display(DECO_RED, zaprefix(taskv[i].t_name), "%s[%s failed]\n", FRC_SPACE, req);
                }
                result++;
                break;
              default :
                if(should_show(SHOW_STATUS)){
                  display(DECO_RED, zaprefix(taskv[i].t_name), "%s[%s failed] (code %d)\n", FRC_SPACE, req, code);
                }
                result++;
                run_shell();
                break;
            }
          } else if (WIFSIGNALED(status)) {
            if(should_show(SHOW_STATUS)){
              display(DECO_RED, zaprefix(taskv[i].t_name), "hit by signal %d\n", WTERMSIG(status));
            }
            if(WTERMSIG(status) != SIGTERM){
              run_shell();
            }
            result++;
          }

          close(taskv[i].t_out);
          taskv[i].t_out = (-1);
          close(taskv[i].t_err);
          taskv[i].t_err = (-1);

          taskv[i].t_pid = (-1);
          taskh--;
        }
      }
    }
  }

  return result;
}

int run_list(char level, char **runv, int runc, struct task *taskv, int taskm, char *req)
{
  int i,j,result;

  result = 0;

  deco_compute(runv, runc);

  i=0;
  j=0;
  while(i<runc){
    switch(test_interactive(level, req, runv[i], STATE_ASK)){
      case CMD_ALT :
        /* if (too many) or (next a different number) run the batch */
        taskv[j++].t_name=runv[i++];
        if((j>=taskm)||(i>=runc)||chprefix(runv[i-1],runv[i])){ 
          result+=run_concurrent(level, taskv, j, req);
          j=0;
        }
        break;
      case CMD_YES :
        taskv[j++].t_name=runv[i++];
        result+=run_concurrent(level, taskv, j, req);
        j=0;
        break;
      case CMD_NO :
        result++;
        i++;
        break;
    }
  }

  if(should_show(SHOW_TOTAL)){
    if(result){
      display_error("failed to %s %d %s of %d in runlevel %c\n", req, result, (result == 1) ? "service" : "services", runc, level);
    } else {
      if(runc){
        display_ok("%s%s %d %s in runlevel %c\n", req, edify(req), runc, (runc == 1) ? "service" : "services", level);
      }
    }
  }

  return result;
}

/****************************************************************************/

static void usage(char *path)
{
  printf("usage: %s [-achlnrsuv] [-b directory] [-d [bceopst]] [-t timeout] runlevel\n", path);
  trms_quit();

  exit(EX_OK);
}

static void copying()
{
  printf("Copyright (C) 2003 Marc Welz. Licensed under the GNU GPL\n");
  trms_quit();

  exit(EX_OK);
}

static void version()
{
  printf("%s %s\n", FRC_NAME, FRC_VERSION);
  trms_quit();

  exit(EX_OK);
}

/****************************************************************************/

int main(int argc, char **argv)
{
  int result;
  char **startv,**stopv, **lastv;
  int startc,stopc,lastc;
  int startf,stopf;
  char runlevel, prevlevel;
  char *env;
  struct task *taskv;
  DIR *dirh;
  struct dirent *dire;
  int show;
  int i,j;
  int taskm;

  /* local configuration stuff */
  int fast = 1;
  int audit = 0;

  /* global configuration stuff */
  frc_login = 0;
  frc_base = NULL;
  frc_setsid = 0;
  frc_state = STATE_OPEN;
  frc_seconds = FRC_TIMEOUT;

  signal(SIGINT, handle_interrupt);
  signal(SIGTERM, SIG_IGN);
  signal(SIGQUIT, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);

  env = getenv("RUNLEVEL");
  runlevel = env ? env[0] : 'N' ;
  env = getenv("PREVLEVEL");
  prevlevel = env ? env[0] : 'N' ;

  result = 0;

  /* deco_setup(SHOW_STDERR | SHOW_STDOUT | SHOW_TRANSIENT | SHOW_STATUS | SHOW_COLOURS); */
  trms_init();
  deco_setup(0xff);

  i=j=1; 
  while(i<argc){
    if(argv[i][0]=='-'){
      switch(argv[i][j]){
        case 'a' :
          audit = 1;
          j++;
          break;
        case 'c' :
          copying();
          j++;
          break;
        case 'h' :
          usage(argv[0]);
          j++;
          break;
        case 'l' :
          frc_login = 0;
          j++;
          break;
        case 'n' :
          frc_state = 0;
          j++;
          break;
        case 'r' :
          frc_state = STATE_OPEN | STATE_AUTH;
          j++;
          break;
        case 's' :
          frc_setsid = 1;
          j++;
          break;
        case 'u' :
          fast = 0;
          j++;
          break;
        case 'v' :
          version();
          j++;
          break;
        case 'b' :
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i >= argc) {
            fatal_error(EX_USAGE, "option -b requires a directory\n");
          }
          frc_base=argv[i] + j;
          i++;
          j = 1;
          break;
        case 't' :
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i >= argc) {
            fatal_error(EX_USAGE, "option -t requires an integer timeout value\n");
          }
          frc_seconds=atoi(argv[i] + j);
          if(frc_seconds <= 0){
            frc_seconds = (-1);
          }
          i++;
          j = 1;
          break;
        case 'd' :
          show = 0;
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i < argc) {
            while(argv[i][j] != '\0'){
              switch(argv[i][j]){
                case 'p' :
                  show |= SHOW_TRANSIENT ; 
                  break;
                case 'o' :
                  show |= SHOW_STDOUT ; 
                  break;
                case 'e' :
                  show |= SHOW_STDERR ; 
                  break;
                case 't' :
                  show |= SHOW_TOTAL ; 
                  break;
                case 's' :
                  show |= SHOW_STATUS ; 
                  break;
                case 'b' :
                  show |= SHOW_BANNER ; 
                  break;
                case 'c' :
                  show |= SHOW_COLOURS ; 
                  break;
                default :
                  display_error("unknown display flag %c\n", argv[i][j]);
                  break;
              }
              j++;
            }
          }
          deco_change(show);
          break;
        case '-' :
          j++;
          break;
        case '\0' :
          j=1;
          i++;
          break;
        default :
          fatal_error(EX_USAGE, "unknown option -%c\n", argv[i][j]);
          break;
      }
    } else if(isdigit(argv[i][0])||(argv[i][0]=='S')){
      runlevel=argv[i][0];
      i++;
      j=1;
    } else {
      fatal_error(EX_USAGE, "unknown runlevel %s\n", argv[i]);
    }
  }

  if(frc_base == NULL){
    frc_base = FRC_BASE;
  } 

  if(!(frc_state & STATE_OPEN)){
    signal(SIGINT, SIG_IGN);
  }

  if(should_show(SHOW_BANNER)){
    display(DECO_YELLOW, NULL, (frc_state & STATE_OPEN) ? "%s %s - press ^C for interactive mode\n" : "%s %s - (c) 2003 Marc Welz\n", FRC_NAME, FRC_VERSION);
  }

  if(runlevel=='N'){
    fatal_error(EX_USAGE, "need to specify a valid runlevel, try -h for help\n");
  }
  
  startv = stopv = lastv = NULL;
  startc = stopc = lastc = 0;

  dirh = runlevel_open(runlevel);
  if(dirh){
    do{
      dire=readdir(dirh);
      if(dire){
#ifdef DEBUG
        fprintf(stderr, "%s\n", dire->d_name);
#endif
        switch(dire->d_name[0]){
          case 'S' :
            startc = vec_insert(&startv, startc, dire->d_name);
            break;
          case 'K' :
            if(prevlevel!='N'){
              stopc = vec_insert(&stopv, stopc, dire->d_name);
            }
            break;
          case '.' :
            break;
          default :
            if(strcmp("README", dire->d_name)){
              display_error("ignoring file %s\n", dire->d_name);
            }
            break;
        }
      }
    }while(dire);
    closedir(dirh);
  }
  
  if(fast){
    if(prevlevel!='N'){
      dirh = runlevel_open(prevlevel);
      if(dirh){
        do{
          dire=readdir(dirh);
          if(dire){
#ifdef DEBUG
            fprintf(stderr, "%s\n", dire->d_name);
#endif
            switch(dire->d_name[0]){
              case 'S' :
                lastc = vec_insert(&lastv, lastc, dire->d_name);
                break;
            }
          }
        }while(dire);
        closedir(dirh);
      }
    }
  }

  taskv = NULL;

  if((startc <= 0) && (stopc <= 0)){
    display_error("nothing to do for change to runlevel %c\n", runlevel);
  } else {

    vec_sort(startv, startc);
    vec_sort(stopv,  stopc);
    vec_sort(lastv,  lastc);

    if(fast){
      /* remove from start list those which (have not been stopped) and (previously started) */
      for(i=0;i<stopc;i++){
        lastc = zadelete(lastv, lastc, stopv[i]);
      }
      for(i=0;i<lastc;i++){
        startc = zadelete(startv, startc, lastv[i]); 
      }
    }

    if(fast){
      taskm = (startc > stopc) ? startc : stopc ;
    } else {
      taskm = 1;
    }

    taskv=malloc(sizeof(struct task) * taskm);
    if(taskv==NULL){
      fatal_error(EX_OSERR, "unable to allocate %d bytes\n", sizeof(struct task) * taskm);
    }

    stopf = run_list(runlevel, stopv,  stopc,  taskv, taskm, REQ_STOP);
    startf = run_list(runlevel, startv, startc, taskv, taskm, REQ_START);

    if(audit){
#ifdef USE_IDSA
      IDSA_CONNECTION *ic;
      ic = idsa_open(FRC_NAME, NULL, 0);
      idsa_set(ic, "run", FRC_NAME, 0, 
          startc ? idsa_risk_make(((double)startf / (double)startc), 0.5) : IDSA_R_NONE, 
          IDSA_R_UNKNOWN, 
          stopf ? IDSA_R_PARTIAL : IDSA_R_MINOR,
          "run-level",    IDSA_T_INT, &runlevel,
          "should-stop",  IDSA_T_INT, &stopc,
          "fail-stop",    IDSA_T_INT, &stopf,
          "should-start", IDSA_T_INT, &startc,
          "fail-start",   IDSA_T_INT, &startf,
          NULL);
      idsa_close(ic);
#else
      openlog(FRC_NAME, 0, LOG_DAEMON);
      syslog((startf || stopf) ? LOG_ERR : LOG_INFO, "run-level=%d should-stop=%d fail-stop=%d should-start=%d fail-start=%d", runlevel, stopc, stopf, startc, startf);
      closelog();
#endif
    }

    if(taskv){
      free(taskv);
      taskm = 0;
    }

#ifdef DEBUG
    vec_dump("start", startv, startc, stderr);
    vec_dump("stop", stopv, stopc, stderr);
    vec_dump("last", lastv, lastc, stderr);
#endif
  }

  trms_quit();

  vec_free(startv, startc);
  vec_free(stopv, stopc);
  vec_free(lastv, lastc);

  return result;
}
